@@ -29,12 +29,14 @@
 #include <stdarg.h>
 #include <string.h>
 #include <assert.h>
+#include <ctype.h>
 #include <libestr.h>
 
 #include "liblognorm.h"
 #include "lognorm.h"
 #include "samp.h"
 #include "ptree.h"
+#include "annot.h"
 #include "internal.h"
 
 /**
@@ -73,6 +75,8 @@ ln_deletePTree(struct ln_ptree *tree)
 	if(tree == NULL)
 		goto done;
 
+	if(tree->tags != NULL)
+		ee_deleteTagbucket(tree->tags);
 	for(node = tree->froot ; node != NULL ; ) {
 		ln_deletePTree(node->subtree);
 		nodeDel = node;
@@ -585,6 +589,77 @@ done:	return r;
 }
 
 
+/**
+ * Special parser for iptables-like name/value pairs.
+ * The pull multiple fields. Note that once this parser has been selected,
+ * it is very unlikely to be left, as it is *very* generic. This parser is
+ * required because practice shows that already-structured data like iptables
+ * can otherwise not be processed by liblognorm in a meaningful way.
+ *
+ * @param[in] tree current tree to process
+ * @param[in] string string to be matched against (the to-be-normalized data)
+ * @param[in/out] offs start position in input data, on exit first unparsed position
+ * @param[in/out] event handle to event that is being created during normalization
+ *
+ * @return 0 if parser was successfully, something else on error
+ */
+static int
+ln_iptablesParser(struct ln_ptree *tree, es_str_t *str, es_size_t *offs,
+		  struct ee_event **event)
+{
+	int r;
+	es_size_t o = *offs;
+	es_str_t *fname;
+	es_str_t *fval;
+	struct ee_value *value;
+	unsigned char *pstr;
+	unsigned char *end;
+
+ln_dbgprintf(tree->ctx, "%d enter iptable parser, len %d", (int) *offs, (int) es_strlen(str));
+	if(o == es_strlen(str)) {
+		r = -1; /* can not be, we have no n/v pairs! */
+		goto done;
+	}
+	
+	end = es_getBufAddr(str) + es_strlen(str);
+	pstr = es_getBufAddr(str) + o;
+	while(pstr < end) {
+		while(isspace(*pstr))
+			++pstr;
+		fname = es_newStr(16);
+		while(!isspace(*pstr) && *pstr != '=') {
+			es_addChar(&fname, *pstr);
+			++pstr;
+		}
+		if(*pstr == '=') {
+			fval = es_newStr(16);
+			++pstr;
+			/* error on space */
+			while(!isspace(*pstr) && pstr < end) {
+				es_addChar(&fval, *pstr);
+				++pstr;
+			}
+		} else {
+			fval = es_newStrFromCStr("[*PRESENT*]", sizeof("[*PRESENT*]")-1);
+		}
+		char *cn, *cv;
+		cn = es_str2cstr(fname, NULL);
+		cv = es_str2cstr(fval, NULL);
+ln_dbgprintf(tree->ctx, "iptable parser extracts %s=%s", cn, cv);
+		value = ee_newValue(tree->ctx->eectx);
+		ee_setStrValue(value, fval);
+		CHKR(addField(tree->ctx, event, fname, value));
+	}
+
+	r = 0;
+	*offs = es_strlen(str);
+
+done:
+	ln_dbgprintf(tree->ctx, "%d iptable parser returns %d", (int) *offs, (int) r);
+	return r;
+}
+
+
 /**
  * Recursive step of the normalizer. It walks the parse tree and calls itself
  * recursively when this is appropriate. It also implements backtracking in
@@ -605,6 +680,7 @@ ln_normalizeRec(struct ln_ptree *tree, es_str_t *str, es_size_t offs, struct ee_
 		struct ln_ptree **endNode)
 {
 	int r;
+	int localR;
 	es_size_t i;
 	int left;
 	ln_fieldList_t *node;
@@ -653,26 +729,53 @@ ln_normalizeRec(struct ln_ptree *tree, es_str_t *str, es_size_t offs, struct ee_
 
 	/* now try the parsers */
 	while(node != NULL) {
-		cstr = es_str2cstr(node->name, NULL);
-		ln_dbgprintf(tree->ctx, "%d:trying parser for field '%s': %p", (int) offs, cstr, node->parser);
-		free(cstr);
+		if(tree->ctx->debug) {
+			cstr = es_str2cstr(node->name, NULL);
+			ln_dbgprintf(tree->ctx, "%d:trying parser for field '%s': %p",
+					(int) offs, cstr, node->parser);
+			free(cstr);
+		}
 		i = offs;
-		if(node->parser(tree->ctx->eectx, str, &i, node->data, &value) == 0) {
-			/* potential hit, need to verify */
-			ln_dbgprintf(tree->ctx, "potential hit, trying subtree");
-			left = ln_normalizeRec(node->subtree, str, i, event, endNode);
-			if(left == 0 && (*endNode)->flags.isTerminal) {
-				ln_dbgprintf(tree->ctx, "%d: parser matches at %d", (int) offs, (int)i);
-				CHKR(addField(tree->ctx, event, node->name, value));
-				r = 0;
-				goto done;
-			} else {
-				ee_deleteValue(value); /* was created, now not needed */
+		if(node->isIPTables) {
+			localR = ln_iptablesParser(tree, str, &i, event);
+			ln_dbgprintf(tree->ctx, "%d iptables parser return, i=%d",
+						(int) offs, (int)i);
+			if(localR == 0) {
+				/* potential hit, need to verify */
+				ln_dbgprintf(tree->ctx, "potential hit, trying subtree");
+				left = ln_normalizeRec(node->subtree, str, i, event, endNode);
+				if(left == 0 && (*endNode)->flags.isTerminal) {
+					ln_dbgprintf(tree->ctx, "%d: parser matches at %d", (int) offs, (int)i);
+					r = 0;
+					goto done;
+				}
+				ln_dbgprintf(tree->ctx, "%d nonmatch, backtracking required, left=%d",
+						(int) offs, (int)left);
+				if(left < r)
+					r = left;
+			}
+		} else {
+			localR = node->parser(tree->ctx->eectx, str, &i, node->data, &value);
+			if(localR == 0) {
+				/* potential hit, need to verify */
+				ln_dbgprintf(tree->ctx, "potential hit, trying subtree");
+				left = ln_normalizeRec(node->subtree, str, i, event, endNode);
+				if(left == 0 && (*endNode)->flags.isTerminal) {
+					ln_dbgprintf(tree->ctx, "%d: parser matches at %d", (int) offs, (int)i);
+					if(!es_strbufcmp(node->name, (unsigned char*)"-", 1))
+						ee_deleteValue(value); /* filler, discard */
+					else
+						CHKR(addField(tree->ctx, event, node->name, value));
+					r = 0;
+					goto done;
+				} else {
+					ee_deleteValue(value); /* was created, now not needed */
+				}
+				ln_dbgprintf(tree->ctx, "%d nonmatch, backtracking required, left=%d",
+						(int) offs, (int)left);
+				if(left < r)
+					r = left;
 			}
-			ln_dbgprintf(tree->ctx, "%d nonmatch, backtracking required, left=%d",
-					(int) offs, (int)left);
-			if(left < r)
-				r = left;
 		}
 		node = node->next;
 	}
@@ -706,8 +809,16 @@ ln_normalize(ln_ctx ctx, es_str_t *str, struct ee_event **event)
 
 	left = ln_normalizeRec(ctx->ptree, str, 0, event, &endNode);
 
-	ln_dbgprintf(ctx, "final result for normalizer: left %d, endNode %p, isTerminal %d, tagbucket %p",
-		     left, endNode, left == 0 ? endNode->flags.isTerminal : 0, endNode->tags);
+	if(ctx->debug) {
+		if(left == 0) {
+			ln_dbgprintf(ctx, "final result for normalizer: left %d, endNode %p, "
+				     "isTerminal %d, tagbucket %p",
+				     left, endNode, endNode->flags.isTerminal, endNode->tags);
+		} else {
+			ln_dbgprintf(ctx, "final result for normalizer: left %d, endNode %p",
+				     left, endNode);
+		}
+	}
 	if(left != 0 || !endNode->flags.isTerminal) {
 		/* we could not successfully parse, some unparsed items left */
 		if(left < 0) {
@@ -718,7 +829,11 @@ ln_normalize(ln_ctx ctx, es_str_t *str, struct ee_event **event)
 	} else {
 		/* success, finalize event */
 		if(endNode->tags != NULL) {
+			if(*event == NULL) {
+				CHKN(*event = ee_newEvent(ctx->eectx));
+			}
 			CHKR(ee_assignTagbucketToEvent(*event, ee_addRefTagbucket(endNode->tags)));
+			CHKR(ln_annotateEvent(ctx, *event));
 		}
 	}
 
